@@ -284,7 +284,7 @@ def test_cast_struct(request: pytest.FixtureRequest, constructor: Constructor) -
         # a column of type MAP<STRING, STRING>
         native_ldf = cast("NativeLazyFrame", native_df)
         _tmp_nw_compliant_frame = nw.from_native(native_ldf)._compliant_frame
-        F = _tmp_nw_compliant_frame._F  # type: ignore[attr-defined] # noqa: N806
+        F = _tmp_nw_compliant_frame._F  # type: ignore[attr-defined]
         T = _tmp_nw_compliant_frame._native_dtypes  # type: ignore[attr-defined] # noqa: N806
 
         native_ldf = native_ldf.withColumn(  # type: ignore[attr-defined]
