@@ -1,6 +1,6 @@
 from __future__ import annotations
 
-from typing import TYPE_CHECKING, Any
+from typing import TYPE_CHECKING, Any, Literal
 
 import pytest
 
@@ -10,7 +10,7 @@ from narwhals.dependencies import get_cudf, get_modin, get_polars
 from tests.utils import POLARS_VERSION, Constructor, assert_equal_data
 
 if TYPE_CHECKING:
-    from types import ModuleType
+    from narwhals._typing import Arrow, Dask, IntoBackend, Modin, Pandas, Polars
 
 
 data = {"a": [1, 2], "b": [3, 4]}
@@ -51,7 +51,7 @@ def test_collect_to_default_backend(constructor: Constructor) -> None:
 )
 @pytest.mark.parametrize("backend", ["pandas", Implementation.PANDAS])
 def test_collect_to_valid_backend_pandas(
-    constructor: Constructor, backend: Implementation | str | None
+    constructor: Constructor, backend: Pandas
 ) -> None:
     pytest.importorskip("pandas")
     import pandas as pd
@@ -66,7 +66,7 @@ def test_collect_to_valid_backend_pandas(
 )
 @pytest.mark.parametrize("backend", ["polars", Implementation.POLARS])
 def test_collect_to_valid_backend_polars(
-    constructor: Constructor, backend: Implementation | str | None
+    constructor: Constructor, backend: Polars
 ) -> None:
     pytest.importorskip("polars")
     import polars as pl
@@ -81,7 +81,7 @@ def test_collect_to_valid_backend_polars(
 )
 @pytest.mark.parametrize("backend", ["pyarrow", Implementation.PYARROW])
 def test_collect_to_valid_backend_pyarrow(
-    constructor: Constructor, backend: Implementation | str | None
+    constructor: Constructor, backend: Arrow
 ) -> None:
     pytest.importorskip("pyarrow")
     import pyarrow as pa
@@ -131,12 +131,12 @@ def test_collect_to_valid_backend_pyarrow_mod(constructor: Constructor) -> None:
     "backend", ["foo", Implementation.DASK, Implementation.MODIN, pytest]
 )
 def test_collect_to_invalid_backend(
-    constructor: Constructor, backend: ModuleType | Implementation | str | None
+    constructor: Constructor, backend: Literal["foo"] | IntoBackend[Modin | Dask]
 ) -> None:
     df = nw.from_native(constructor(data))
 
     with pytest.raises(ValueError, match="Unsupported `backend` value"):
-        df.lazy().collect(backend=backend).to_native()
+        df.lazy().collect(backend=backend).to_native()  # type: ignore[arg-type]
 
 
 def test_collect_with_kwargs(constructor: Constructor) -> None:
