@@ -1,8 +1,7 @@
 from __future__ import annotations
 
 import narwhals as nw
-from tests.utils import ConstructorEager
-from tests.utils import assert_equal_data
+from tests.utils import ConstructorEager, assert_equal_data
 
 
 def test_is_duplicated(constructor_eager: ConstructorEager) -> None:
